package org.warp.jcwdb.database;

import org.apache.commons.lang3.reflect.FieldUtils;
import org.warp.jcwdb.EnhancedObject;
import org.warp.jcwdb.annotations.*;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.LinkedHashMap;
import java.util.Map;

public class DatabaseDataInitializer implements IDataInitializer {

	private final DatabaseObjectsIO objectsIO;

	public DatabaseDataInitializer(DatabaseObjectsIO objectsIO) {
		this.objectsIO = objectsIO;
	}

	@Override
	public void initializeDBObject(EnhancedObject obj) throws IOException {
		initializeDBObjectFields(obj);
		initializeDBObjectPrimitiveFields(obj);
		initializeDBObjectProperties(obj);
	}

	private void initializeDBObjectFields(EnhancedObject obj) throws IOException {
		// Declare the variables needed to get the biggest field Id
		Field[] unorderedFields = objectsIO.getFields(obj);
		// Find the biggest field Id
		int biggestFieldId = objectsIO.getBiggestFieldId(unorderedFields);

		// Allocate new UIDs
		long[] fieldUIDs = objectsIO.allocateNewUIDs(biggestFieldId + 1);

		// Declare the other variables
		Field[] fields = new Field[biggestFieldId + 1];
		DBDataType[] orderedFieldTypes = new DBDataType[biggestFieldId + 1];

		// Load all fields metadata and load them
		for (Field field : unorderedFields) {
			DBField fieldAnnotation = field.getAnnotation(DBField.class);
			int fieldId = fieldAnnotation.id();
			DBDataType fieldType = fieldAnnotation.type();
			objectsIO.loadField(obj, field, fieldType, fieldUIDs[fieldId]);
			fields[fieldId] = field;
			orderedFieldTypes[fieldId] = fieldType;
		}
		// Set fields metadata
		obj.setFields(fields, orderedFieldTypes, fieldUIDs);
	}

	private void initializeDBObjectPrimitiveFields(EnhancedObject obj) throws IOException {
		// Declare the variables needed to get the biggest field Id
		Field[] unorderedFields = objectsIO.getPrimitiveFields(obj);
		// Find the biggest field Id
		int biggestFieldId = objectsIO.getBiggestPrimitiveFieldId(unorderedFields);

		// Allocate new UID
		long fieldDataUID = objectsIO.newNullObject();

		// Declare the other variables
		Field[] fields = new Field[biggestFieldId + 1];
		DBPrimitiveType[] orderedFieldTypes = new DBPrimitiveType[biggestFieldId + 1];

		// Load all fields metadata and load them
		try {
			for (Field field : unorderedFields) {
				DBPrimitiveField fieldAnnotation = field.getAnnotation(DBPrimitiveField.class);
				int fieldId = fieldAnnotation.id();
				DBPrimitiveType fieldType = fieldAnnotation.type();
				switch (fieldType) {
					case BOOLEAN:
						FieldUtils.writeField(field, obj, false, true);
						break;
					case BYTE:
						FieldUtils.writeField(field, obj, (byte) 0, true);
						break;
					case CHAR:
						FieldUtils.writeField(field, obj, (char) 0, true);
						break;
					case SHORT:
						FieldUtils.writeField(field, obj, (short) 0, true);
						break;
					case INTEGER:
						FieldUtils.writeField(field, obj, 0, true);
						break;
					case LONG:
						FieldUtils.writeField(field, obj, (long) 0, true);
						break;
					case FLOAT:
						FieldUtils.writeField(field, obj, (float) 0, true);
						break;
					case DOUBLE:
						FieldUtils.writeField(field, obj, (double) 0, true);
						break;
				}
				fields[fieldId] = field;
				orderedFieldTypes[fieldId] = fieldType;
			}
		} catch (IllegalArgumentException | IllegalAccessException e) {
			throw new IOException(e);
		}
		// Set fields metadata
		obj.setPrimitiveFields(fields, orderedFieldTypes, fieldDataUID);
	}

	private void initializeDBObjectProperties(EnhancedObject obj) throws IOException {
		// Declare the variables needed to get the biggest property Id
		Method[] unorderedPropertyGetters = obj.getPropertyGetters();
		Method[] unorderedPropertySetters = obj.getPropertySetters();

		// Find the biggest property Id
		int biggestGetter = objectsIO.getBiggestPropertyGetterId(unorderedPropertyGetters);
		int biggestSetter = objectsIO.getBiggestPropertySetterId(unorderedPropertySetters);
		int biggestPropertyId = biggestGetter > biggestSetter ? biggestGetter : biggestSetter;

		// Allocate new UIDs
		long[] propertyUIDs = objectsIO.allocateNewUIDs(biggestPropertyId + 1);

		for (Method property : unorderedPropertySetters) {
			DBPropertySetter fieldAnnotation = property.getAnnotation(DBPropertySetter.class);
			int propertyId = fieldAnnotation.id();
			if (propertyId > biggestPropertyId) {
				biggestPropertyId = propertyId;
			}
		}

		// Declare the other variables
		DBDataType[] propertyTypes = new DBDataType[biggestPropertyId + 1];
		Method[] propertyGetters = new Method[biggestPropertyId + 1];
		Method[] propertySetters = new Method[biggestPropertyId + 1];
		Map<String, DBPropertySetter> setterMethods = new LinkedHashMap<>();
		Map<String, DBPropertyGetter> getterMethods = new LinkedHashMap<>();

		// Load the properties metadata
		for (Method property : unorderedPropertyGetters) {
			DBPropertyGetter propertyAnnotation = property.getAnnotation(DBPropertyGetter.class);
			int propertyId = propertyAnnotation.id();
			DBDataType propertyType = propertyAnnotation.type();
			propertyTypes[propertyId] = propertyType;
			propertyGetters[propertyId] = property;
			getterMethods.put(property.getName(), propertyAnnotation);
		}
		for (Method property : unorderedPropertySetters) {
			DBPropertySetter propertyAnnotation = property.getAnnotation(DBPropertySetter.class);
			int propertyId = propertyAnnotation.id();
			DBDataType propertyType = propertyAnnotation.type();
			propertyTypes[propertyId] = propertyType;
			propertySetters[propertyId] = property;
			setterMethods.put(property.getName(), propertyAnnotation);
		}
		// Set properties metadata
		obj.setProperties(propertyGetters, propertySetters, propertyTypes, propertyUIDs, setterMethods, getterMethods);
	}
}
